//   Copyright (c) 2010 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include <iostream>
#include <sstream>
#include <exception>
#include <cassert>
#include <vector>
#include <cstdlib>
#include <memory>
#include <boost/scope_exit.hpp>
#include <boost/foreach.hpp>
#include <boost/algorithm/string/case_conv.hpp>
#include <boost/iostreams/device/file.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/newline.hpp>
#if 0
#include <zip.h>
#endif

#include "CoCoA/library.H"
#include "Interpreter.H"

using namespace std;
using namespace boost;
using namespace boost::iostreams;
using namespace CoCoA::AST;
using namespace CoCoA::LexerNS;
using namespace CoCoA::ParserNS;

namespace CoCoA {
namespace InterpreterNS {

namespace {

typedef pair<string, intrusive_ptr<BuiltInFunction> > NameFunPair;
vector<NameFunPair> builtIns;

struct AddBuiltIn {
	AddBuiltIn(string fname, intrusive_ptr<BuiltInFunction> fun) {
		builtIns.push_back(make_pair(fname, fun));
	}
};

#define DECLARE_ARITYCHECK_FUNCTION(FNAME) bool builtin_check_arity_##FNAME(int nArg)
#define DECLARE_BUILTIN_FUNCTION(FNAME) \
		intrusive_ptr<RightValue> builtin_##FNAME(const intrusive_ptr<const InvocationExpression> invocationExpression, RuntimeEnvironment * const runtimeEnv); \
		DECLARE_ARITYCHECK_FUNCTION(FNAME) ;\
		AddBuiltIn add_##FNAME(#FNAME, new BuiltInFunction(builtin_##FNAME, builtin_check_arity_##FNAME)); \
		intrusive_ptr<RightValue> builtin_##FNAME(const intrusive_ptr<const InvocationExpression> invocationExpression, RuntimeEnvironment * const runtimeEnv)
#define DECLARE_STD_BUILTIN_FUNCTION(FNAME, NARG) \
	DECLARE_ARITYCHECK_FUNCTION(FNAME) { return nArg==(NARG); } \
	DECLARE_BUILTIN_FUNCTION(FNAME) { \
	invocationExpression->checkNumberOfArgs(NARG);
#define END_STD_BUILTIN_FUNCTION }
#define EVAL_ZZValue(NARG) runtimeEnv->evalArgAs<ZZValue>(invocationExpression->args[NARG-1])->theInteger

template <typename T1, typename T2> const T2 &theValue(intrusive_ptr<T1> x);
template <> const ZZ &theValue<ZZValue, ZZ>(intrusive_ptr<ZZValue> x) { return x->theInteger; }
template <> const QQ &theValue<QQValue, QQ>(intrusive_ptr<QQValue> x) { return x->theRational; }
template <> const RingElem &theValue<PolyValue, RingElem>(intrusive_ptr<PolyValue> x) { return x->thePoly; }
template <> const ring &theValue<RingValue, ring>(intrusive_ptr<RingValue> x) { return x->theRing; }

template <typename T> struct CoCoALibType {};
template <> struct CoCoALibType<QQValue> { typedef QQ type; };
template <> struct CoCoALibType<ZZValue> { typedef ZZ type; };
template <> struct CoCoALibType<PolyValue> { typedef RingElem type; };
template <> struct CoCoALibType<RingValue> { typedef ring type; };

#define DECLARE_COCOALIB_FUNCTION1(RETT, FUN, ARG1T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 1) \
	return new RETT(FUN(theValue<ARG1T, CoCoALibType<ARG1T>::type>(runtimeEnv->evalArgAs<ARG1T>(invocationExpression->args[0])))); }

#define DECLARE_COCOALIB_BOOL_FUNCTION1(FUN, ARG1T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 1) \
	return BoolValue::fromBool(FUN(theValue<ARG1T, CoCoALibType<ARG1T>::type>(runtimeEnv->evalArgAs<ARG1T>(invocationExpression->args[0])))); }

#define DECLARE_COCOALIB_BOOL_FUNCTION2(FUN, ARG1T, ARG2T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 2) \
	return BoolValue::fromBool(FUN( \
				theValue<ARG1T, CoCoALibType<ARG1T>::type>(runtimeEnv->evalArgAs<ARG1T>(invocationExpression->args[0])), \
				theValue<ARG2T, CoCoALibType<ARG2T>::type>(runtimeEnv->evalArgAs<ARG2T>(invocationExpression->args[1]))  \
				)); }

#define DECLARE_COCOALIB_FUNCTION2(RETT, FUN, ARG1T, ARG2T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 2) \
	return new RETT(FUN( \
			theValue<ARG1T, CoCoALibType<ARG1T>::type>(runtimeEnv->evalArgAs<ARG1T>(invocationExpression->args[0])), \
			theValue<ARG2T, CoCoALibType<ARG2T>::type>(runtimeEnv->evalArgAs<ARG2T>(invocationExpression->args[1]))  \
			)); }

#define DECLARE_COCOALIB_FUNCTION3(RETT, FUN, ARG1T, ARG2T, ARG3T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 3) \
	return new RETT(FUN( \
			theValue<ARG1T, CoCoALibType<ARG1T>::type>(runtimeEnv->evalArgAs<ARG1T>(invocationExpression->args[0])), \
			theValue<ARG2T, CoCoALibType<ARG2T>::type>(runtimeEnv->evalArgAs<ARG2T>(invocationExpression->args[1])), \
			theValue<ARG3T, CoCoALibType<ARG3T>::type>(runtimeEnv->evalArgAs<ARG3T>(invocationExpression->args[2]))  \
			)); }

// functions on ZZ
DECLARE_COCOALIB_FUNCTION1(ZZValue, factorial, ZZValue)
DECLARE_COCOALIB_FUNCTION1(ZZValue, fibonacci, ZZValue)
DECLARE_COCOALIB_FUNCTION2(ZZValue, binomial, ZZValue, ZZValue)
DECLARE_COCOALIB_FUNCTION2(ZZValue, random, ZZValue, ZZValue)
DECLARE_COCOALIB_FUNCTION3(ZZValue, PowerMod, ZZValue, ZZValue, ZZValue)
DECLARE_COCOALIB_BOOL_FUNCTION1(IsEven, ZZValue) // AMB
DECLARE_COCOALIB_BOOL_FUNCTION1(IsProbPrime, ZZValue) // AMB
// functions on RingElem
DECLARE_COCOALIB_BOOL_FUNCTION1(IsInvertible, PolyValue)
DECLARE_COCOALIB_BOOL_FUNCTION1(IsConstant, PolyValue) // AMB
DECLARE_COCOALIB_BOOL_FUNCTION1(IsIndet, PolyValue) // AMB
DECLARE_COCOALIB_BOOL_FUNCTION1(IsMonomial, PolyValue) // AMB
DECLARE_COCOALIB_BOOL_FUNCTION1(IsHomog, PolyValue) // AMB
DECLARE_COCOALIB_BOOL_FUNCTION2(IsDivisible, PolyValue, PolyValue)
DECLARE_COCOALIB_FUNCTION2(PolyValue, gcd, PolyValue, PolyValue)
DECLARE_COCOALIB_FUNCTION2(PolyValue, deriv, PolyValue, PolyValue) // AMB
DECLARE_COCOALIB_FUNCTION1(PolyValue, LC, PolyValue) // AMB
DECLARE_COCOALIB_FUNCTION1(PolyValue, CommonDenom, PolyValue) // AMB
DECLARE_COCOALIB_FUNCTION1(PolyValue, content, PolyValue) // AMB
DECLARE_COCOALIB_FUNCTION2(PolyValue, power, PolyValue, ZZValue)
DECLARE_COCOALIB_FUNCTION1(ZZValue, characteristic, RingValue)
// functions on ring/PolyRing
DECLARE_COCOALIB_BOOL_FUNCTION1(IsCommutative, RingValue)
DECLARE_COCOALIB_BOOL_FUNCTION1(IsIntegralDomain, RingValue)
DECLARE_COCOALIB_BOOL_FUNCTION1(IsGCDDomain, RingValue)
DECLARE_COCOALIB_BOOL_FUNCTION1(IsOrderedDomain, RingValue)
DECLARE_COCOALIB_BOOL_FUNCTION1(IsField, RingValue)
DECLARE_COCOALIB_BOOL_FUNCTION1(IsZ, RingValue)
DECLARE_COCOALIB_BOOL_FUNCTION1(IsFractionField, RingValue)
DECLARE_COCOALIB_BOOL_FUNCTION1(IsPolyRing, RingValue)
DECLARE_COCOALIB_BOOL_FUNCTION1(IsQuotientRing, RingValue)
DECLARE_COCOALIB_FUNCTION1(PolyValue, zero, RingValue)
DECLARE_COCOALIB_FUNCTION1(PolyValue, one, RingValue)
DECLARE_COCOALIB_FUNCTION1(RingValue, owner, PolyValue)
DECLARE_COCOALIB_FUNCTION1(RingValue, NewFractionField, RingValue)  // AMB

DECLARE_STD_BUILTIN_FUNCTION(IsInteger, 2) {
	intrusive_ptr<LeftValue> resultRef = intrusive_ptr_cast<LeftValue>(runtimeEnv->evalArg(invocationExpression->args.front(), RuntimeEnvironment::EVAL_BY_REF));
	intrusive_ptr<PolyValue> poly = runtimeEnv->evalArgAs<PolyValue>(invocationExpression->args[1]);
	ZZ zz;
	bool isInteger(IsInteger(zz, poly->thePoly));
	if (isInteger) {
		if (resultRef->assignmentNeedsOwnership())
			resultRef->obtainOwnership();
		resultRef->assign(new ZZValue(zz));
	}
	return BoolValue::fromBool(isInteger);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsRational, 2) {
	intrusive_ptr<LeftValue> resultRef = intrusive_ptr_cast<LeftValue>(runtimeEnv->evalArg(invocationExpression->args.front(), RuntimeEnvironment::EVAL_BY_REF));
	intrusive_ptr<PolyValue> poly = runtimeEnv->evalArgAs<PolyValue>(invocationExpression->args[1]);
	QQ qq;
	bool isRational(IsRational(qq, poly->thePoly));
	if (isRational) {
		if (resultRef->assignmentNeedsOwnership())
			resultRef->obtainOwnership();
		resultRef->assign(new QQValue(qq));
	}
	return BoolValue::fromBool(isRational);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(aliases, 0) {
	runtimeEnv->topLevelAliases->dump(runtimeEnv->getOutputStream());
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(pause, 0) {
#ifdef C5IDE
	runtimeEnv->interpreter->singleStepExecution = true;
#else // #ifdef C5IDE
	(void)runtimeEnv; // keeps the compiler happy (or, at least, silent ;-) )
#endif // #ifdef C5IDE
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

#ifdef C5IDE
DECLARE_STD_BUILTIN_FUNCTION(sleep, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<ZZValue> zz = runtimeEnv->evalArgAs<ZZValue>(arg);
	long l;
	if (!IsConvertible(l, zz->theInteger) || l<0 || l>1000)
		throw RuntimeException("The argument must be in the range [0..1000]", arg.exp);
	this_thread::sleep(boost::posix_time::seconds(static_cast<int>(l)));
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(debug, 1) {
	const Argument &arg = invocationExpression->args.front();
	string filename = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	try {
		runtimeEnv->interpreter->singleStepExecution = true;
		runtimeEnv->interpreter->readAndExecute(filename, false);
	} catch (const RuntimeException &) {
		throw;
	} catch (const BaseException &be) {
		throw RuntimeException(be.reason, invocationExpression->getBegin(), invocationExpression->getEnd());
	}
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION
#endif // #ifdef C5IDE

#if 0
DECLARE_STD_BUILTIN_FUNCTION(ZipFileList, 1) {
	const Argument &arg = invocationExpression->args.front();
	const string filename = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	int err;
	struct zip *zf = ::zip_open(filename.c_str(), ZIP_CHECKCONS, &err);
	if (!zf)
		throw RuntimeException("Cannot read Zip file \""+filename+"\"", arg.exp);
	intrusive_ptr<ListValue> result(new ListValue);
	const int numEntries = zip_get_num_files(zf);
	for(int a=0; a<numEntries; ++a)
		result->addValue(new StringValue(zip_get_name(zf, a, ZIP_FL_UNCHANGED)));
	::zip_close(zf);
	return result;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ZipRead, 2) {
	const Argument &argFilename = invocationExpression->args.front();
	const Argument &argEntryName = invocationExpression->args[1];
	const string filename = runtimeEnv->evalArgAs<StringValue>(argFilename)->theString;
	const string entryName = runtimeEnv->evalArgAs<StringValue>(argEntryName)->theString;
	int err;
	struct zip *zf = ::zip_open(filename.c_str(), ZIP_CHECKCONS, &err);
	if (!zf)
		throw RuntimeException("Cannot read Zip file \""+filename+"\"", argFilename.exp);
	const int index = zip_name_locate(zf, entryName.c_str(), /* flags */ 0);
	if (index<0) {
		::zip_close(zf);
		throw RuntimeException("Cannot find an entry named \""+entryName+" inside the Zip file \""+filename+"\"", argEntryName.exp);
	}
	struct zip_file *entry = ::zip_fopen_index(zf, index, /* flags */ 0);
	if (!entry) {
		::zip_close(zf);
		throw RuntimeException("Error reading the entry named \""+entryName+" inside the Zip file \""+filename+"\"", argEntryName.exp);
	}
	string result;
	for(;;) {
		char buf[1024];
		int nRead = ::zip_fread(entry, buf, sizeof(buf));
		if (nRead<=0) {
			::zip_fclose(entry);
			::zip_close(zf);
			return new StringValue(result);
		}
		result.append(buf, buf+nRead);
	}
}
END_STD_BUILTIN_FUNCTION
#endif

DECLARE_STD_BUILTIN_FUNCTION(len, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<RightValue> v = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(arg, RuntimeEnvironment::EVAL_BY_VALUE));
	if (intrusive_ptr<StringValue> s = dynamic_pointer_cast<StringValue>(v))
		return new ZZValue(ZZ(s->theString.length()));
	if (intrusive_ptr<ListValue> list = dynamic_pointer_cast<ListValue>(v))
		return new ZZValue(ZZ(list->size()));
	throw WrongTypeException(ListValue::type->name + " or " + StringValue::type->name, v->getType()->name, arg.exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(sum,1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(arg);
	const ListValue::ContainerType::size_type size = list->size();
	intrusive_ptr<RightValue> result = ZZValue::zero;
	for(ListValue::ContainerType::size_type a=0; a<size; ++a)
	try	{
			result = runtimeEnv->binaryOperatorDispatch(result, list->getValue(a), RuntimeEnvironment::opPlusMap, arg.exp->getBegin(), arg.exp->getEnd());
	} catch (const InterruptException &) {
		throw;
	} catch (const RuntimeException &) {
		throw RuntimeException("Some elements have incompatible types for the sum", arg.exp);
	}
	return result;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(product, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(arg);
	const ListValue::ContainerType::size_type size = list->size();
	intrusive_ptr<RightValue> result = ZZValue::one;
	try	{
		for(ListValue::ContainerType::size_type a=0; a<size; ++a)
			result = runtimeEnv->binaryOperatorDispatch(result, list->getValue(a), RuntimeEnvironment::opStarMap, arg.exp->getBegin(), arg.exp->getEnd());
	} catch (const InterruptException &) {
		throw;
	} catch (const RuntimeException &) {
		throw RuntimeException("Some elements have incompatible types for the product", arg.exp);
	}
	return result;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ScalarProduct, 2) {
	const Argument &arg1 = invocationExpression->args.front();
	intrusive_ptr<ListValue> l1 = runtimeEnv->evalArgAs<ListValue>(arg1);
	const Argument &arg2 = invocationExpression->args[1];
	intrusive_ptr<ListValue> l2 = runtimeEnv->evalArgAs<ListValue>(arg2);
	const ListValue::ContainerType::size_type size = l1->size();
	const CharPointer &begin = arg1.exp->getBegin();
	const CharPointer &end = arg2.exp->getEnd();
	if (l2->size()!=size)
		throw RuntimeException("The arguments must be lists of the same size", begin, end);
	intrusive_ptr<RightValue> result = ZZValue::zero;
	try	{
		for(ListValue::ContainerType::size_type a=0; a<size; ++a)
			result = runtimeEnv->binaryOperatorDispatch(
						result,
						runtimeEnv->binaryOperatorDispatch(l1->getValue(a), l2->getValue(a), RuntimeEnvironment::opStarMap, begin, end),
						RuntimeEnvironment::opPlusMap,
						begin,
						end);
	} catch (const InterruptException &) {
		throw;
	} catch (const RuntimeException &) {
		throw RuntimeException("Some elements have incompatible types for the scalar product", begin, end);
	}
	return result;
}
END_STD_BUILTIN_FUNCTION

DECLARE_ARITYCHECK_FUNCTION(concat) { (void)nArg; return true; }

DECLARE_BUILTIN_FUNCTION(concat) {
	intrusive_ptr<ListValue> returnValue = new ListValue();
	BOOST_FOREACH(const Argument &arg, invocationExpression->args) {
		intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(arg);
		int size = list->size();
		for(int a=0; a<size; ++a)
			returnValue->addValue(list->getValue(a));
	}
	return returnValue;
}

DECLARE_STD_BUILTIN_FUNCTION(append, 2) {
	intrusive_ptr<RightValue> elem = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(invocationExpression->args[1], RuntimeEnvironment::EVAL_BY_VALUE));
	intrusive_ptr<ListValue> list = runtimeEnv->obtainUnsharedList(invocationExpression->args.front());
	list->addValue(elem);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(remove, 2) {
	const Argument indexArg = invocationExpression->args[1];
	intrusive_ptr<RightValue> shouldBeAnIndex = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(indexArg, RuntimeEnvironment::EVAL_BY_VALUE));
	intrusive_ptr<ListValue> list = runtimeEnv->obtainUnsharedList(invocationExpression->args.front());
	ListValue::ContainerType::size_type index = list->checkIndex(shouldBeAnIndex, indexArg.exp);
	list->removeValue(index);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(insert, 3) {
	const Argument &indexArg = invocationExpression->args[1];
	intrusive_ptr<RightValue> shouldBeAnIndex = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(indexArg, RuntimeEnvironment::EVAL_BY_VALUE));
	intrusive_ptr<RightValue> elem = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(invocationExpression->args[2], RuntimeEnvironment::EVAL_BY_VALUE));
	intrusive_ptr<ListValue> list = runtimeEnv->obtainUnsharedList(invocationExpression->args.front());
	ListValue::ContainerType::size_type index = list->checkIndex(shouldBeAnIndex, indexArg.exp);
	list->insertValue(index, elem);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(div, 2) {
	const Argument &arg1 = invocationExpression->args[0];
	intrusive_ptr<ZZValue> a = runtimeEnv->evalArgAs<ZZValue>(arg1);
	const Argument &arg2 = invocationExpression->args[1];
	intrusive_ptr<ZZValue> b = runtimeEnv->evalArgAs<ZZValue>(arg2);
	ZZ result;
	div(result, a->theInteger, b->theInteger);
	return new ZZValue(result);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(mod, 2) {
	const Argument &arg1 = invocationExpression->args[0];
	intrusive_ptr<ZZValue> a = runtimeEnv->evalArgAs<ZZValue>(arg1);
	const Argument &arg2 = invocationExpression->args[1];
	intrusive_ptr<ZZValue> b = runtimeEnv->evalArgAs<ZZValue>(arg2);
	ZZ result;
	mod(result, a->theInteger, b->theInteger);
	return new ZZValue(result);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(NumDigits, 2) {
	const Argument &arg1 = invocationExpression->args[0];
	intrusive_ptr<ZZValue> a = runtimeEnv->evalArgAs<ZZValue>(arg1);
	const Argument &arg2 = invocationExpression->args[1];
	intrusive_ptr<ZZValue> b = runtimeEnv->evalArgAs<ZZValue>(arg2);
	long base;
	if (!IsConvertible(base, b->theInteger) || base<2 || base>36)
		throw RuntimeException("Base must be in the range 2..36", arg2.exp);
	return new ZZValue(ZZ(NumDigits(a->theInteger, base)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(sign, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<const RightValue> v(intrusive_ptr_cast<const RightValue>(runtimeEnv->evalArg(arg, RuntimeEnvironment::EVAL_BY_VALUE)));
	if (const intrusive_ptr<const PolyValue> poly = dynamic_pointer_cast<const PolyValue>(v))
		return ZZValue::fromInt(sign(poly->thePoly));
	if (const intrusive_ptr<const ZZValue> zz = dynamic_pointer_cast<const ZZValue>(v))
		return ZZValue::fromInt(sign(zz->theInteger));
	if (const intrusive_ptr<const QQValue> qq = dynamic_pointer_cast<const QQValue>(v))
		return ZZValue::fromInt(sign(qq->theRational));
	throw WrongTypeException(ZZValue::type->name+", "+QQValue::type->name+" or "+PolyValue::type->name, v->getType()->name, arg.exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsZero, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<const RightValue> v(intrusive_ptr_cast<const RightValue>(runtimeEnv->evalArg(arg, RuntimeEnvironment::EVAL_BY_VALUE)));
	if (const intrusive_ptr<const PolyValue> poly = dynamic_pointer_cast<const PolyValue>(v))
		return BoolValue::fromBool(IsZero(poly->thePoly));
	if (const intrusive_ptr<const ZZValue> zz = dynamic_pointer_cast<const ZZValue>(v))
		return BoolValue::fromBool(IsZero(zz->theInteger));
	if (const intrusive_ptr<const QQValue> qq = dynamic_pointer_cast<const QQValue>(v))
		return BoolValue::fromBool(IsZero(qq->theRational));
	throw WrongTypeException(ZZValue::type->name+", "+QQValue::type->name+" or "+PolyValue::type->name, v->getType()->name, arg.exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsOne, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<const RightValue> v(intrusive_ptr_cast<const RightValue>(runtimeEnv->evalArg(arg, RuntimeEnvironment::EVAL_BY_VALUE)));
	if (const intrusive_ptr<const PolyValue> poly = dynamic_pointer_cast<const PolyValue>(v))
		return BoolValue::fromBool(IsOne(poly->thePoly));
	if (const intrusive_ptr<const ZZValue> zz = dynamic_pointer_cast<const ZZValue>(v))
		return BoolValue::fromBool(IsOne(zz->theInteger));
	if (const intrusive_ptr<const QQValue> qq = dynamic_pointer_cast<const QQValue>(v))
		return BoolValue::fromBool(IsOne(qq->theRational));
	throw WrongTypeException(ZZValue::type->name+", "+QQValue::type->name+" or "+PolyValue::type->name, v->getType()->name, arg.exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsMinusOne, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<const RightValue> v(intrusive_ptr_cast<const RightValue>(runtimeEnv->evalArg(arg, RuntimeEnvironment::EVAL_BY_VALUE)));
	if (const intrusive_ptr<const PolyValue> poly = dynamic_pointer_cast<const PolyValue>(v))
		return BoolValue::fromBool(IsMinusOne(poly->thePoly));
	if (const intrusive_ptr<const ZZValue> zz = dynamic_pointer_cast<const ZZValue>(v))
		return BoolValue::fromBool(IsMinusOne(zz->theInteger));
	if (const intrusive_ptr<const QQValue> qq = dynamic_pointer_cast<const QQValue>(v))
		return BoolValue::fromBool(IsMinusOne(qq->theRational));
	throw WrongTypeException(ZZValue::type->name+", "+QQValue::type->name+" or "+PolyValue::type->name, v->getType()->name, arg.exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(abs, 1) {
	const Argument &arg = invocationExpression->args.front();
	bool isZZ;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<ZZValue, QQValue>(arg, isZZ);
	if (isZZ)
		return new ZZValue(abs(intrusive_ptr_cast<ZZValue>(v)->theInteger));
	return new QQValue(abs(intrusive_ptr_cast<QQValue>(v)->theRational));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(floor, 1) {
	const Argument &arg = invocationExpression->args.front();
	bool isZZ;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<ZZValue, QQValue>(arg, isZZ);
	if (isZZ)
		return v;
	return new ZZValue(floor(intrusive_ptr_cast<QQValue>(v)->theRational));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ceil, 1) {
	const Argument &arg = invocationExpression->args.front();
	bool isZZ;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<ZZValue, QQValue>(arg, isZZ);
	if (isZZ)
		return v;
	return new ZZValue(ceil(intrusive_ptr_cast<QQValue>(v)->theRational));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(round, 1) {
	const Argument &arg = invocationExpression->args.front();
	bool isZZ;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<ZZValue, QQValue>(arg, isZZ);
	if (isZZ)
		return v;
	return new ZZValue(round(intrusive_ptr_cast<QQValue>(v)->theRational));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(fields, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<RecordValue> r = runtimeEnv->evalArgAs<RecordValue>(arg);
	return r->fieldNames();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ConcatList, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<ListValue> listOfList = runtimeEnv->evalArgAs<ListValue>(arg);
	intrusive_ptr<ListValue> result(new ListValue);
	const ListValue::ContainerType::size_type nLists = listOfList->size();
	for(ListValue::ContainerType::size_type a=0; a<nLists; ++a) {
		intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(listOfList->getValue(a));
		if (!l)
			throw RuntimeException("The argument is not a list of lists", arg.exp);
		const ListValue::ContainerType::size_type len = l->size();
		for(ListValue::ContainerType::size_type b=0; b<len; ++b)
			result->addValue(l->getValue(b));
	}
	return result;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(num, 1) {
	const Argument &arg = invocationExpression->args.front();
	bool isZZ;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<ZZValue, QQValue>(arg, isZZ);
	if (isZZ)
		return v;
	return new ZZValue(num(intrusive_ptr_cast<QQValue>(v)->theRational));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(den, 1) {
	const Argument &arg = invocationExpression->args.front();
	bool isZZ;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<ZZValue, QQValue>(arg, isZZ);
	if (isZZ)
		return ZZValue::one;
	return new ZZValue(den(intrusive_ptr_cast<QQValue>(v)->theRational));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(sprint, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<RightValue> obj = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(arg, RuntimeEnvironment::EVAL_BY_VALUE));
	intrusive_ptr<OutputStringStreamValue> outs(new OutputStringStreamValue);
	outs->print(runtimeEnv, obj, arg.exp);
	return outs->close(runtimeEnv, invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(GetErrMesg, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<ErrorValue> err = runtimeEnv->evalArgAs<ErrorValue>(arg);
	return new StringValue(err->message);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(close, 1) {
	const Argument &arg = invocationExpression->args.front();
	return runtimeEnv->evalArgAs<OutputStreamValue>(arg)->close(runtimeEnv, invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(OpenOutputString, 0) {
	(void)runtimeEnv; // to avoid a compilation warning
	return new OutputStringStreamValue();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(GetEnv, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<StringValue> varName = runtimeEnv->evalArgAs<StringValue>(arg);
	char *value = ::getenv(varName->theString.c_str());
	if (!value)
		return StringValue::empty;
	return new StringValue(string(value));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(OpenOFile, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<StringValue> filename = runtimeEnv->evalArgAs<StringValue>(arg);
	return new OutputFileStreamValue(filename, arg.exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ContentsOfFile, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<StringValue> filename = runtimeEnv->evalArgAs<StringValue>(arg);
	filtering_istream input;
	input.push(newline_filter(newline::posix));
	const file_source fs(filename->theString.c_str());
	if (!fs.is_open())
		throw RuntimeException("Cannot open file \""+filename->theString+"\" for reading.", invocationExpression);
	input.push(fs);
	string wholeFile;
	for(;;) {
		string line;
		getline(input, line);
		if (input.bad())
			throw RuntimeException("Cannot read from \""+filename->theString+"\".", invocationExpression);
		wholeFile += line;
		if (input.eof())
			break;
		wholeFile += '\n';
	}
	return new StringValue(wholeFile);
}
END_STD_BUILTIN_FUNCTION

const string invalidTagError("The tag must be a valid identifier");

DECLARE_STD_BUILTIN_FUNCTION(tagged, 2) {
	const Argument &argValue = invocationExpression->args.front();
	const Argument &argTag = invocationExpression->args[1];
	intrusive_ptr<RightValue> v = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(argValue, RuntimeEnvironment::EVAL_BY_VALUE));
	const string tag = runtimeEnv->evalArgAs<StringValue>(argTag)->theString;
	if (!LexerNS::Lexer::isValidIdentifier(tag))
		throw RuntimeException(invalidTagError, argTag.exp);
	return new TaggedValue(v->untagged(), invocationExpression->packageName+"."+tag);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(untagged, 1) {
	const Argument &argValue = invocationExpression->args.front();
	intrusive_ptr<RightValue> v = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(argValue, RuntimeEnvironment::EVAL_BY_VALUE));
	return v->untagged();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(tag, 1) {
	const Argument &argValue = invocationExpression->args.front();
	intrusive_ptr<RightValue> v = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(argValue, RuntimeEnvironment::EVAL_BY_VALUE));
	if (intrusive_ptr<TaggedValue> taggedValue = dynamic_pointer_cast<TaggedValue>(v)) {
		const string::size_type l = invocationExpression->packageName.length()+1;
		if (taggedValue->tag.substr(0, l)==invocationExpression->packageName+".")
			return new StringValue(taggedValue->tag.substr(l, string::npos));
		return new StringValue(taggedValue->tag);
	}
	return StringValue::empty;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(type, 1) {
	const Argument &argValue = invocationExpression->args.front();
	intrusive_ptr<RightValue> v = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(argValue, RuntimeEnvironment::EVAL_BY_VALUE));
	return v->getType();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(PkgName, 0) {
	(void)runtimeEnv; // to avoid a compilation warning
	return new StringValue(invocationExpression->packageName);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(/* uppercase because it's a backward-compatibility hack */ Functions, 1) {
	(void)runtimeEnv; // to avoid a compilation warning
	throw RuntimeException("Use the statement Describe to obtain package information", invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(TAGGED, 1) {
	const Argument &arg = invocationExpression->args.front();
	const string tag = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	if (!LexerNS::Lexer::isValidIdentifier(tag))
		throw RuntimeException(invalidTagError, arg.exp);
	return TypeValue::tagType(tag);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(types, 0) {
	(void)runtimeEnv; // to avoid a compilation warning
	throw RuntimeException("Cannot return a list of all types because they're infinite (CoCoA 4 lies about this, I don't)", invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(error, 1) {
	const Argument &arg = invocationExpression->args.front();
	const string message = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	throw RuntimeException(message, invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ascii, 1) {
	const Argument &arg = invocationExpression->args.front();
	intrusive_ptr<RightValue> v = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(arg, RuntimeEnvironment::EVAL_BY_VALUE));
	if (intrusive_ptr<StringValue> sv = dynamic_pointer_cast<StringValue>(v)) {
		intrusive_ptr<ListValue> l(new ListValue());
		const string s(sv->theString);
		BOOST_FOREACH(const char c, s)
			l->addValue(new ZZValue(ZZ(static_cast<int>(c))));
		return l;
	}
	if (intrusive_ptr<ZZValue> zz = dynamic_pointer_cast<ZZValue>(v)) {
		long l;
		if (!IsConvertible(l, zz->theInteger) || l<0 || l>255)
			throw RuntimeException("When the argument is an integer, it must be in the range [0..255]", arg.exp);
		return new StringValue(string(1, static_cast<char>(l)));
	}
	if (intrusive_ptr<ListValue> list = dynamic_pointer_cast<ListValue>(v)) {
		string retValue;
		ListValue::ContainerType::size_type size = list->size();
		for(ListValue::ContainerType::size_type a=0; a<size; ++a) {
			intrusive_ptr<ZZValue> zz = dynamic_pointer_cast<ZZValue>(list->getValue(a));
			long l;
			if (!zz || !IsConvertible(l, zz->theInteger) || l<0 || l>255)
				throw RuntimeException("When the argument is a list, it must be a list of integers only in the range [0..255]", arg.exp);
			retValue += static_cast<char>(l);
		}
		return new StringValue(retValue);
	}
	throw WrongTypeException(ListValue::type->name+", "+StringValue::type->name + " or "+ZZValue::type->name, v->getType()->name, arg.exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(LT, 1) { // AMB
	const Argument &arg = invocationExpression->args.front();
	return new PolyValue(LT_forC5(runtimeEnv->evalArgAs<PolyValue>(arg)->thePoly));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(LPP, 1) { // AMB
	const Argument &arg = invocationExpression->args.front();
	return new PolyValue(LPP_forC5(runtimeEnv->evalArgAs<PolyValue>(arg)->thePoly));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(LM, 1) { // AMB
	const Argument &arg = invocationExpression->args.front();
	return new PolyValue(LM_forC5(runtimeEnv->evalArgAs<PolyValue>(arg)->thePoly));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(monomials, 1) { // AMB
	const Argument &arg = invocationExpression->args.front();
	return new ListValue(monomials_forC5(runtimeEnv->evalArgAs<PolyValue>(arg)->thePoly));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(support, 1) { // AMB
	const Argument &arg = invocationExpression->args.front();
	return new ListValue(support_forC5(runtimeEnv->evalArgAs<PolyValue>(arg)->thePoly));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(log, 1) { // AMB
	const Argument &arg = invocationExpression->args.front();
  intrusive_ptr<PolyValue> f = runtimeEnv->evalArgAs<PolyValue>(arg);
  vector<ZZ> expv;
  ZZExponents(expv, LPP(f->thePoly));
	return new ListValue(expv);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(indets, 1) { // AMB
	const Argument &arg = invocationExpression->args.front();
	return new ListValue(indets(AsPolyRing((runtimeEnv->evalArgAs<RingValue>(arg))->theRing)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IndetsCalled, 2) { // AMB
	const Argument &arg1 = invocationExpression->args[0];
	const Argument &arg2 = invocationExpression->args[1];
	return new ListValue(IndetsCalled((runtimeEnv->evalArgAs<StringValue>(arg1))->theString,
                                     AsPolyRing((runtimeEnv->evalArgAs<RingValue>(arg2))->theRing)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(indet, 2) { // AMB
	const Argument &arg1 = invocationExpression->args[0];
	intrusive_ptr<RingValue> a = runtimeEnv->evalArgAs<RingValue>(arg1);
	const Argument &arg2 = invocationExpression->args[1];
	intrusive_ptr<ZZValue> b = runtimeEnv->evalArgAs<ZZValue>(arg2);
	return new PolyValue(indet(AsPolyRing(a->theRing), b->theInteger-1));
}
END_STD_BUILTIN_FUNCTION

} // namespace {

void RuntimeEnvironment::initBuiltInFunctions()
{
	BOOST_FOREACH(NameFunPair &p, builtIns)
		this->setTopLevelVar(p.first, p.second, VariableSlot::VSF_SystemProtected);
}

} // namespace AST
} // namespace CoCoA
