//   Copyright (c)  2010  Anna Bigatti

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include "OnlineHelp.H"
#include "CoCoA/library.H"

#include <algorithm>
using std::transform;
#include <cctype>
//using std::tolower;
#include <fstream>
using std::ifstream;
#include <iostream>
using std::cout;
using std::endl;
#include <string>
using std::string;
#include <vector>
using std::vector;


extern string packageDir;

namespace CoCoA
{
namespace OnlineHelp
{

  // fwd decl -- defined later in this file
  std::string LowerCase(const std::string& str);
  bool IsStr1BeforeStr2(std::ifstream& in, const std::string& s1, const std::string& s2);
  std::string StringInTag(const std::string& line, const std::string& XMLTag);
  std::string SkipToStringInTag(std::ifstream& in, const std::string& XMLTag);

  inline bool IsSubstring(std::string s, std::string line) 
  { return line.find(s) != string::npos; }

  //-- index --

  index::index()
  {
    CoCoA::GlobalLogput() << "Loading index ..." << std::flush;
    ifstream in(XMLFileName());
    if (in == 0)
      CoCoA_ERROR("Cannot find input file `CoCoAHelp.xml`", "OnlineHelp::index()");
    string line; // used only inside while loop
    while (!in.eof())
    {
      getline(in, line);
      if (IsSubstring("<!--", line)) continue;
      if (IsSubstring("<command>", line))
        myAddEntry(in, SkipToStringInTag(in, "<title>"), "</command>");
      if (IsSubstring("<section>", line))
        myAddEntry(in, SkipToStringInTag(in, "<title>"), "</section>");
    }
    CoCoA::GlobalLogput() << "... index done" << endl;
  }


  void index::myAddEntry(std::ifstream& in, const std::string& title, const std::string& ClosingTag)
  {
    myVec.push_back(entry(title));
    entry& e(myVec.back());
    e.myAddKey(title);
    if (!IsStr1BeforeStr2(in, "<keys>", ClosingTag)) return;
    string line;
    getline(in, line);
    string s; // used only inside while loop
    while ((!in.eof()) && (!IsSubstring("</keys>",line)))
    {
      if ((s=StringInTag(line, "<key>")) != "") e.myAddKey(s);
      getline(in, line);
    }
  }


  const index& UniqueIndex();


  //-- index::entry --
  void index::entry::myAddKey(const std::string& key)
  {
    myKeysValue.push_back(LowerCase(key));
  }


  //----------------------------------------------------------------

  std::string LowerCase(const std::string& str)
  {
    string s(str); // slightly wasteful, but readable
    transform(s.begin(), s.end(), s.begin(), tolower);
    return s;
  }


  bool IsStr1BeforeStr2(std::ifstream& in, const std::string& s1, const std::string& s2)
  {
    string line;
    getline(in, line);
    while (!in.eof())
    {
      if ( IsSubstring(s2, line) ) return false;
      if ( IsSubstring(s1, line) ) return true;
      getline(in, line);
    }
    CoCoA_ERROR(CoCoA::ERR::SERIOUS, "IsStr1BeforeStr2");
    return false;
  }
  

  // only for opening and closing tag in the same line
  std::string StringInTag(const std::string& line, const std::string& XMLTag)
  {
    size_t open;
    if ( (open=line.find(XMLTag)) == string::npos) return "";
    
    string ClosedXMLTag = XMLTag;
    ClosedXMLTag.replace(0, 1, "</");
    size_t close = line.find(ClosedXMLTag);
    if (close == string::npos)
      CoCoA_ERROR(XMLTag+" closing tag not found in this line", "StringInTag");
    size_t StartPos = open + XMLTag.length();
    return line.substr(StartPos, close-StartPos);
  }


  std::string SkipToStringInTag(std::ifstream& in, const std::string& XMLTag)
  {
    string line;
    getline(in, line);
    string s;
    while (!in.eof())
    {
      if ( (s=StringInTag(line, XMLTag)) != "") return s;
      getline(in, line);
    }
    return "";
  }
  

  void ReplaceWithQuotes(std::string& line, std::string s)
  {
    size_t open;
    while ( (open=line.find(s)) != string::npos)
      line.replace(open, s.length(), "\"");
  }
  

  void PrintExample(std::ifstream& in, std::ostream &out)
  {
    out << "//==================  EXAMPLE ===================\\\\" << endl;
    string line;
    getline(in, line);
    while ( !IsSubstring("</example>", line) )
    {
      ReplaceWithQuotes(line, "<quotes>");
      ReplaceWithQuotes(line, "</quotes>");
      out << line << endl;
      getline(in, line);
    }
    out << "\\\\==================  o=o=o=o ===================//" << endl;
  }
  

  void ProcessLine(std::string& line)
  {
    ReplaceWithQuotes(line, "<quotes>");
    ReplaceWithQuotes(line, "</quotes>");
    ReplaceWithQuotes(line, "<tt>");
    ReplaceWithQuotes(line, "</tt>");
    ReplaceWithQuotes(line, "<ttref>");
    ReplaceWithQuotes(line, "</ttref>");
    ReplaceWithQuotes(line, "<ref>");
    ReplaceWithQuotes(line, "</ref>");
    ReplaceWithQuotes(line, "<em>");
    ReplaceWithQuotes(line, "</em>");
    if (IsSubstring("<par/>", line)) line = "";
  }
  

void PrintSearchExact(std::string title, std::ostream &out)
{
  ifstream in(XMLFileName());
  if (in == 0)
  {
    std::cerr << "Cannot find input file `CoCoAHelp.xml`.  Aborting." << endl;
    abort();
  }
  string s = SkipToStringInTag(in, "<title>");
  while (!in.eof() && s != title)  s = SkipToStringInTag(in, "<title>");
  if (in.eof())
    CoCoA_ERROR("Not found exact title: " + title, "PrintSearchExact");
  out << "==============[ " << title << " ]===============" << endl;
  string line;
  getline(in, line);
  while (!IsSubstring("<description>", line))  getline(in, line);
  out << "== DESCRIPTION ==" << endl;
  getline(in, line);
  while (!IsSubstring("</description>", line))
  {
    if ( IsSubstring("<example>", line))  PrintExample(in, out);
    else
    {
      ProcessLine(line);
      out << line << endl;
    }
    getline(in, line);
  }
}


#if 0 
  // hopefully useless
  std::size_t SkipToLAngle(std::ifstream& in, std::string& line, std::size_t from)
  {
    size_t pos;
    while (!in.eof())
    {
      if ( (pos=line.find("<")) != string::npos) return pos;
      getline(in, line);
    }
    return string::npos;
  }
#endif


//----------------------------------------------------------------------

void SearchMatch(std::vector<std::string>& MatchingTitles, const std::string& s)
{
  string ls(LowerCase(s));
  
  const index& index(UniqueIndex());
  for (size_t i=0; i<index.mySize(); ++i)
    for (size_t j=0; j<index.myKeys(i).size(); ++j)
      if (IsSubstring(ls, index.myKeys(i)[j]))
      {
        MatchingTitles.push_back(index.myTitle(i));
        break;
      }
}


void PrintMan(const std::string& s, std::ostream &out)
{
  vector<string> found;
  SearchMatch(found, s);
  if (found.size() == 1)  PrintSearchExact(found[0], out);
  else
    for (size_t i=0; i<found.size(); ++i)
      if (LowerCase(found[i]) == LowerCase(s))
        PrintSearchExact(found[i], out);
  if (found.size() == 0)
    out << "\nNo matches for \"" << s << "\"" << endl;
  else
  {
    out << "\nAll matches for \"" << s << "\":" << endl;
    for (size_t i=0; i<found.size(); ++i)
      out << " ? " << found[i] << endl;
    out << endl;
  }
}


const char* XMLFileName()
{
  static const string UniqueCopy(packageDir+"/man/CoCoAHelp.xml");
  return UniqueCopy.c_str();
}


const index& UniqueIndex()
{
  static index UniqueCopy;
  return UniqueCopy;
}

} // namespace OnlineHelp
} // namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/CoCoA-5/OnlineHelp.C,v 1.10 2010/11/22 17:45:02 bigatti Exp $
// $Log: OnlineHelp.C,v $
// Revision 1.10  2010/11/22 17:45:02  bigatti
// -- fixed bug in storing keys
//
// Revision 1.9  2010/11/22 17:37:44  bigatti
// -- improved message for entries with no matching
//
// Revision 1.8  2010/09/02 13:04:08  bigatti
// -- added conversion for <em>
//
// Revision 1.7  2010/09/01 13:24:48  bigatti
// -- moved all manual functions into CoCoA::OnlineHelp namespace
//
// Revision 1.6  2010/09/01 12:27:52  lagorio
// *** empty log message ***
//
// Revision 1.5  2010/09/01 09:23:19  bigatti
// -- added some "const"
// -- using CoCoA::GlobalLogput and CoCoA_ERROR instead of clog
//
// Revision 1.4  2010/09/01 08:28:35  lagorio
// *** empty log message ***
//
// Revision 1.3  2010/09/01 07:46:19  lagorio
// *** empty log message ***
//
// Revision 1.2  2010/08/31 14:55:02  bigatti
// -- pretty basic fixes
//

